#!/usr/bin/env python3
#
# Copyright (c) 2015-2018, Fabian Greif
# Copyright (c) 2018, Niklas Hauser
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

import sys
import random
import logging
import collections

import lbuild.module
import lbuild.environment

from .exception import LbuildException
from .exception import LbuildBuildException
from .exception import LbuildOptionFormatException

from .node import BaseNode, NameResolver
from .config import ConfigNode
from .facade import BuildLogFacade

from . import repository
from . import utils
from . import config

LOGGER = logging.getLogger('lbuild.parser')

class Runner:
    def __init__(self, node, env):
        self.node = node
        self.env = env

    def validate(self):
        if hasattr(self.node, "validate"):
            self.node.validate(self.env.facade_validate)

    def build(self):
        self.node.build(self.env.facade_build)

    def post_build(self, buildlog):
        if hasattr(self.node, "post_build"):
            self.node.post_build(self.env.facade_post_build, BuildLogFacade(buildlog))


class Parser(BaseNode):
    def __init__(self, config=None):
        BaseNode.__init__(self, "lbuild", BaseNode.Type.PARSER)
        self._config = config if config else ConfigNode()
        self._config_flat = self._config.flatten()

    @property
    def config(self):
        return self._config_flat

    @property
    def modules(self):
        return {m.fullname:m for m in self.all_modules()}

    @property
    def repositories(self):
        return {r.fullname:r for r in self._findall(BaseNode.Type.REPOSITORY)}

    def load_repositories(self, repofilenames=None):
        repofiles = set(utils.listify(repofilenames))
        parsed = set()
        config_map = {}

        while True:
            # flatten configuration and collect repositories
            self._config_flat = self._config.flatten()
            repofiles = (repofiles | set(self._config_flat.repositories)) - parsed
            if not (len(repofiles) + len(parsed)):
                LOGGER.error("\n" + str(self._config.render()))
                raise LbuildException("No repositories loaded!")
            parsed |= repofiles

            # Parse only new repositories
            for repofile in repofiles:
                config_map.update(self.parse_repository(repofile)._config_map)

            # nothing more to extend
            if not len(self._config_flat._extends):
                break

            for filename, aliases in self._config_flat._extends.items():
                node = self._config.find(filename)
                for alias in aliases:
                    if alias not in config_map:
                        raise LbuildException("Configuration alias '{}' not found in any map! "
                                              "Available aliases: '{}'".format(alias, "', '".join(config_map)))
                    self._config.extend(node, ConfigNode.from_file(config_map[alias]))
                del node._extends[filename]

        self._update_format();
        LOGGER.info("\n" + str(self._config.render()))
        return self._config_flat

    def parse_repository(self, repofilename: str) -> repository.Repository:
        """
        Parse the given repository file.

        Executes the 'prepare' function to populate the repository
        structure.
        """
        repo = repository.Repository.parse_repository(repofilename)
        if repo.name in [c.name for c in self.children]:
            raise LbuildException("Repository name '{}' is ambiguous. Name must be unique.".format(repo.name))
        repo.parent = self

        return repo

    def merge_repository_options(self):
        # only deal with repo options that contain one `:`
        resolver = self.option_resolver
        for name, value in {n:v for n,v in self.config.options.items() if n.count(":") == 1}.items():
            try:
                resolver[name].value = value
            except LbuildException as e:
                raise LbuildException("Failed to merge repository options!\n" + str(e))
        # return all repo options in tree
        return {o.fullname:o for o in self.all_options(depth=3)}

    def prepare_repositories(self, repo_options):
        """
        Prepare and select modules which are available given the set of
        repository repo_options.

        Returns:
            dict: Available modules, key is the qualified module name.
        """
        undefined = self._undefined_options(repo_options)
        if len(undefined):
            raise LbuildException("Unknown values for options '{}'. Please provide a value in the "
                                  "configuration file or on the command line.".format("', '".join(undefined)))
        modules = []
        for repo in self._findall(BaseNode.Type.REPOSITORY):
            modules.extend(repo.prepare(repo_options))
        modules = lbuild.module.build_modules(modules)
        if len(modules) == 0:
            raise LbuildBuildException("No module found with the selected repository options!")

        self._resolve_dependencies(ignore_failure=True)
        self._update_format()
        return modules

    def merge_module_options(self):
        # only deal with repo options that contain one `:`
        resolver = self.option_resolver
        for name, value in {n:v for n,v in self.config.options.items() if n.count(":") > 1}.items():
            try:
                resolver[name].value = value
            except LbuildException as e:
                raise LbuildException("Failed to merge module options!\n" + str(e))
        # return all module options in tree
        return {o.fullname:o for o in self.all_options() if o.depth > 2}

    def resolve_dependencies(self, requested_modules, depth=sys.maxsize):
        # map the dependency names to the node objects
        self._resolve_dependencies()
        return self._filter_dependencies(requested_modules, depth)

    def _filter_dependencies(self, requested_modules, depth=sys.maxsize):
        """
        Resolve dependencies by adding missing modules.

        Args:
            modules (list): All modules available in the system
            requested_modules (list): Selected modules. Sub-set of all modules.
            depth (int): Maximum depth of dependencies. If not set to maximum
                not all dependencies might be resolved.

        Returns:
            list: Required modules for the given list of modules.
        """
        selected_modules = requested_modules.copy()

        LOGGER.info("Selected modules: {}".format(
                    ", ".join(sorted([module.fullname for module in selected_modules]))))

        current = selected_modules
        while depth > 0:
            additional = []
            for module in current:
                for dependency in module.dependencies:
                    if dependency not in selected_modules and dependency not in additional:
                        LOGGER.debug("Add dependency: " + dependency.fullname)
                        additional.append(dependency)
            if not additional:
                # Abort if no new dependencies are being found
                break
            selected_modules.extend(additional)
            current = additional
            additional = []
            depth -= 1

        # disable all non-selected modules
        for module in self.all_modules():
            if module._available:
                module._available = module in selected_modules
                # if not module._available:
                #     LOGGER.debug("Disabled: " + module.fullname)
        self._update();

        return selected_modules

    def find_module(self, name):
        return self.module_resolver[name]

    def find_option(self, name):
        return self.option_resolver[name]

    def find(self, name):
        if ":" not in name:
            if name not in self.repositories:
                raise LbuildException("No repository found for '{}'!".format(name))
            return self.repositories[name]

        try:
            return self.find_module(name)
        except LbuildException as e:
            if not "but searching for" in str(e):
                raise e
        return self.find_option(name)

    def find_modules(self, queries):
        return self.find_any(queries, self.Type.MODULE)

    def find_any(self, queries, types=None):
        nodes = set()
        for query in utils.listify(queries):
            nodes |= set(self._resolve_partial(query, set()))
        if types:
            types = lbuild.utils.listify(types)
            nodes = filter(lambda n: any(n.type == t for t in types), nodes)
        return list(nodes)

    @staticmethod
    def _undefined_options(options):
        undefined = []
        for fullname, option in options.items():
            if option.value is None:
                undefined.append(fullname)
        return undefined

    @staticmethod
    def validate_modules(build_modules):
        Parser.build_modules(build_modules, None)

    @staticmethod
    def build_modules(build_modules, buildlog):
        if not len(build_modules):
            raise LbuildException("No modules selected, aborting!")

        groups = collections.defaultdict(list)
        for node in (build_modules + list(set(m.repository for m in build_modules))):
            env = lbuild.environment.Environment(node, buildlog)
            groups[node.depth].append(Runner(node, env))

        exceptions = []
        # Enforce that the submodules are always build before their
        # parent modules.
        for index in sorted(groups, reverse=True):
            group = groups[index]
            random.shuffle(group)

            for runner in group:
                try:
                    runner.validate()
                except lbuild.exception.LbuildValidateException as error:
                    exceptions.append(error)

        if len(exceptions) > 0:
            raise lbuild.exception.LbuildAggregateException(exceptions)

        # Cannot build with these settings, just pre_build
        if buildlog is None:
            return

        for index in sorted(groups, reverse=True):
            group = groups[index]
            random.shuffle(group)

            for runner in group:
                runner.build()

        for index in sorted(groups, reverse=True):
            group = groups[index]
            random.shuffle(group)

            for runner in group:
                runner.post_build(buildlog)




#!/usr/bin/env python3
#
# Copyright (c) 2015-2018, Fabian Greif
# Copyright (c) 2018, Niklas Hauser
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

import os
import time
import shutil
import fnmatch
import jinja2
import logging

import lbuild.filter
import lbuild.utils

from .facade import EnvironmentValidateFacade, EnvironmentBuildFacade, EnvironmentPostBuildFacade
from .exception import LbuildException, LbuildTemplateException, LbuildForwardException

simulate = False


def _copyfile(sourcepath, destpath):
    """
    Copy a file if the source file time stamp is newer than the destination
    timestamp.
    """
    if not os.path.exists(destpath):
        if not simulate:
            shutil.copy2(sourcepath, destpath)
    else:
        time_diff = os.stat(sourcepath).st_mtime - os.stat(destpath).st_mtime
        if time_diff > 1:
            print(destpath, "override")
            if not simulate:
                shutil.copy2(sourcepath, destpath)


def _copytree(logger, src, dst, ignore=None):
    """
    Implementation of shutil.copytree that overwrites files instead
    of aborting.
    """
    files = os.listdir(src)
    if ignore is not None:
        ignored = ignore(src, files)
    else:
        ignored = set()

    for filename in files:
        if filename not in ignored:
            sourcepath = os.path.join(src, filename)
            destpath = os.path.join(dst, filename)
            if os.path.isdir(sourcepath):
                _copytree(logger, sourcepath, destpath, ignore)
            else:
                starttime = time.time()
                if not os.path.exists(dst):
                    os.makedirs(dst)
                _copyfile(sourcepath, destpath)
                endtime = time.time()
                total = endtime - starttime
                logger(sourcepath, destpath, total)


class Environment:

    def __init__(self, module, buildlog):
        self.options = module.option_value_resolver
        self.modules = module.module_resolver
        self.__module = module
        self.__modulepath = module._filepath
        self.__repopath = module.repository._filepath
        self.__outpath = buildlog.outpath

        self.__buildlog = buildlog
        self.__template_environment = None
        self.__template_environment_filters = None
        self.__template_global_substitutions = {
            'time': time.strftime("%d %b %Y, %H:%M:%S", time.localtime()),
            'options': self.options,
        }

        self.log = logging.getLogger("user." + module.fullname.replace(":", "."))
        self.outbasepath = None
        self.substitutions = {}

    @property
    def facade_validate(self):
        return EnvironmentValidateFacade(self)

    @property
    def facade_build(self):
        return EnvironmentBuildFacade(self)

    @property
    def facade_post_build(self):
        return EnvironmentPostBuildFacade(self)

    def copy(self, src, dest=None, ignore=None):
        """
        Copy file or directory from the modulepath to the buildpath.

        If src or dest is a relative path it will be relocated to the
        module/buildpath. Absolute paths are not changed.

        If dest is empty the same name as src is used (relocated to
        the output path).
        """
        def wrap_ignore(path, files):
            ignored = lbuild.utils.ignore_patterns(*self.__module._ignore_patterns)(path, files)
            ignored.add(self.__module._filename)
            if ignore:
                ignored |= set(ignore(path, files))
            return ignored

        if dest is None:
            dest = src

        starttime = time.time()

        srcpath = os.path.normpath(src if os.path.isabs(src) else self.modulepath(src))
        destpath = os.path.normpath(dest if os.path.isabs(dest) else self.outpath(dest))

        srcrelpath = os.path.relpath(srcpath, self.__repopath)
        if srcrelpath.startswith(".."):
            raise LbuildException("Cannot access files outside of the repository!\n"
                                "'{}'".format(srcrelpath))

        if os.path.isdir(srcpath):
            _copytree(lambda src, dest, time: self.__buildlog.log(self.__module, src, dest, time),
                      srcpath,
                      destpath,
                      wrap_ignore)
        else:
            if not os.path.exists(os.path.dirname(destpath)):
                os.makedirs(os.path.dirname(destpath))
            _copyfile(srcpath, destpath)

            endtime = time.time()
            total = endtime - starttime
            self.__buildlog.log(self.__module, srcpath, destpath, total)

    def __reload_template_environment(self, filters):
        if self.__template_environment_filters == filters:
            return

        self.__template_environment_filters = filters

        # Overwrite jinja2 Environment in order to enable relative paths
        # since this runs locally that should not be a security concern
        # Code from:
        # http://stackoverflow.com/questions/8512677/how-to-include-a-template-with-relative-path-in-jinja2
        class RelEnvironment(jinja2.Environment):
            """
            Override join_path() to enable relative template paths.

            Take care of paths. Jinja seems to use '/' as path separator in
            templates.
            """

            def join_path(self, template, parent):
                path = os.path.join(os.path.dirname(parent), template)
                return os.path.normpath(path).replace('\\','/')

        environment = RelEnvironment(loader=jinja2.FileSystemLoader(self.__repopath),
                                     extensions=['jinja2.ext.do'],
                                     undefined=jinja2.StrictUndefined)

        environment.filters.update(self.__module._filters)
        environment.filters.update(filters)

        # Jinja2 Line Statements
        environment.line_statement_prefix = '%%'
        environment.line_comment_prefix = '%#'

        self.__template_environment = environment

    @property
    def template_environment(self):
        if self.__template_environment is None:
            self.__reload_template_environment({})
        return self.__template_environment

    def template(self, src, dest=None, substitutions=None, filters=None):
        """
        Uses the Jinja2 template engine to generate files.

        If dest is empty the same name as src is used (relocated to
        the output path).
        """
        starttime = time.time()

        if dest is None:
            # If src ends in ".in" remove that and use the remaing part
            # as new destination.
            parts = src.split(".")
            if parts[-1] == "in":
                dest = ".".join(parts[:-1])
            else:
                dest = src

        src = self.repopath(src)
        if src.startswith(".."):
            raise LbuildException("Cannot access template outside of repository!\n"
                                "'{}'".format(src))

        if substitutions is None:
            substitutions = {}

        substitutions.update(self.substitutions)
        try:
            if filters is not None:
                # Reload environment if it uses different filters than
                # the previous environment
                self.__reload_template_environment(filters)

            template = self.template_environment.get_template(src.replace('\\','/'), globals=self.__template_global_substitutions)
            output = template.render(substitutions)
        except jinja2.TemplateNotFound as error:
            raise LbuildException('Failed to retrieve Template: %s' % error)
        except (jinja2.exceptions.TemplateAssertionError,
                jinja2.exceptions.TemplateSyntaxError) as error:
            raise LbuildException("Error in template '{}:{}':\n"
                                " {}: {}".format(error.filename,
                                                 error.lineno,
                                                 error.__class__.__name__,
                                                 error))
        except jinja2.exceptions.UndefinedError as error:
            raise LbuildTemplateException("Error in template '{}':\n"
                                        " {}: {}".format(self.modulepath(src),
                                                         error.__class__.__name__,
                                                         error))
        except LbuildException as error:
            raise LbuildException("Error in template '{}': \n"
                                "{}".format(self.modulepath(src), error))
        except Exception as error:
            raise LbuildForwardException("Error in template '{}': \n"
                                       "{}".format(self.modulepath(src), error),
                                       error)

        outfile_name = self.outpath(dest)

        # Create folder structure if it doesn't exists
        if not simulate:
            if not os.path.exists(os.path.dirname(outfile_name)):
                os.makedirs(os.path.dirname(outfile_name))

            with open(outfile_name, 'w') as outfile:
                outfile.write(output)

        endtime = time.time()
        total = endtime - starttime
        self.__buildlog.log(self.__module, self.modulepath(src), outfile_name, total)

    def modulepath(self, *path):
        """Relocate given path to the path of the module file."""
        return os.path.join(self.__modulepath, *path)

    def repopath(self, *path):
        """Relocate given path to the path of the repo file."""
        return os.path.relpath(self.modulepath(*path), self.__repopath)

    def outpath(self, *path, basepath=None):
        """Relocate given path to the output path."""
        if basepath is None:
            basepath = self.outbasepath
        if basepath is None:
            return os.path.join(self.__outpath, *path)
        else:
            return os.path.join(self.__outpath, basepath, *path)

    def generated_local_files(self, filterfunc=None):
        """
        Get all files which have been generated by this module and its
        submodules so far.

        Keyword arguments:
        filterfunc -- Function which is called to determine whether a path
            should be added to the output. If the function returns `True`
            the path is added.
        """
        filenames = []
        operations = self.__buildlog.operations_per_module(self.__module.fullname)
        for operation in operations:
            filename = os.path.normpath(os.path.join(os.path.relpath(os.path.dirname(operation.filename_out),
                                                                     self.outpath('.')),
                                                     os.path.basename(operation.filename_out)))
            if filterfunc is None or filterfunc(filename):
                filenames.append(filename)
        return filenames

    def append_metadata(self, key, value):
        """
        Append additional information to the build log which can be used in the
        post-build step to generate additional files/data.
        """
        self.__buildlog.metadata[key].append(value)

    def append_metadata_unique(self, key, value):
        """
        Append additional information build log if it is not already present.

        See also `append_metadata`.
        """
        if value not in self.__buildlog.metadata[key]:
            self.__buildlog.metadata[key].append(value)

    def has_option(self, key):
        """Query whether an option exists."""
        try:
            _ = self.options[key]
            return True
        except LbuildException:
            return False

    def has_module(self, key):
        """Query whether a module exists."""
        try:
            _ = self.modules[key]
            return True
        except LbuildException:
            return False

    def get_option(self, key, default=None):
        """
        Get an option value.

        Returns a user configurable default value if the option is not found.
        """
        try:
            return self.options[key]
        except LbuildException:
            return default

    def __getitem__(self, key):
        """
        Get an option value.

        Raises an exception if the option is not found.
        """
        return self.options[key]

    def __repr__(self):
        return repr(self.options)

    def __len__(self):
        return len(self.options)




#!/usr/bin/env python3
#
# Copyright (c) 2015-2018, Fabian Greif
# Copyright (c) 2018, Niklas Hauser
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

import os
import sys
import enum
import shutil
import anytree
import logging
import textwrap
import colorful
import itertools

import lbuild.filter
from lbuild.format import format_node, format_description, format_short_description

from .exception import LbuildException, LbuildAttributeException

LOGGER = logging.getLogger('lbuild.node')


def load_functions_from_file(repository, filename: str, required, optional=None, local={}):
    try:
        localpath = os.path.dirname(os.path.realpath(filename))
        local.update({
            # The localpath(...) function can be used to create
            # a local path form the folder of the repository file.
            'localpath': RelocatePath(localpath),
            'repopath': RelocatePath(repository._filepath),
            'FileReader': LocalFileReaderFactory(localpath),
            'listify': lbuild.filter.listify,
            # 'ignore_patterns': shutil.ignore_patterns,

            'StringOption': lbuild.option.StringOption,
            'BooleanOption': lbuild.option.BooleanOption,
            'NumericOption': lbuild.option.NumericOption,
            'EnumerationOption': lbuild.option.EnumerationOption,
            'SetOption': lbuild.option.SetOption,
        })

        # LOGGER.debug("Parse filename '%s'", filename)
        local = lbuild.utils.with_forward_exception(repository,
                lambda: lbuild.utils.load_module_from_file(filename, local))
        functions = lbuild.utils.get_global_functions(local, required, optional)
        return functions

    except FileNotFoundError as error:
        raise LbuildException("Repository configuration file not found '{}'.".format(filename))

    except KeyError as error:
        raise LbuildException("Invalid repository configuration file '{}':\n"
                            " {}: {}".format(filename,
                                             error.__class__.__name__,
                                             error))

class RelocatePath:
    def __init__(self, basepath):
        self.basepath = basepath
    def __call__(self, *args):
        return os.path.join(self.basepath, *args)

class LocalFileReader:
    def __init__(self, basepath, filename):
        self.basepath = basepath
        self.filename = filename
    def __str__(self):
        return self.read()
    def read(self):
        with open(os.path.join(self.basepath, self.filename)) as file:
            return file.read()

class LocalFileReaderFactory:
    def __init__(self, basepath):
        self.basepath = basepath
    def __call__(self, filename):
        return LocalFileReader(self.basepath, filename)

class Renderer(anytree.RenderTree):
    def __init__(self, node):
        anytree.RenderTree.__init__(self, node,
                                    style=anytree.ContRoundStyle(),
                                    childiter=self.childsort)

    def __str__(self):
        lines = [pre + format_node(node, pre) for pre, _, node in self]
        return "\n".join(lines)

    @staticmethod
    def childsort(items):
        def sorting(item):
            return (item._type != BaseNode.Type.OPTION, item.name)
        return sorted(items, key=sorting)


class NameResolver:
    """
    Name resolver for node.
    """

    def __init__(self, node, nodetype):
        self._node = node
        self._type = nodetype
        self._str = nodetype.name.lower()
        self._value_resolver = False

    def __getitem__(self, key: str):
        node = self._node._resolve_partial_max(key, max_results=1)[0]
        if not node._available:
            raise LbuildException("{} '{}' is not available!".format(self._str, node.fullname))
        if node._type != self._type:
            raise LbuildException("'{}' is of type '{}', but searching for '{}'!".format(
                                  node.fullname, node._type.name.lower(), self._str))
        if node._type == BaseNode.Type.OPTION and self._value_resolver:
            return node.value
        return node

    def get(self, key, default=None):
        try:
            return self.__getitem__(key)
        except:
            return default

    def __contains__(self, key):
        try:
            _ = self.__getitem__(key)
            return True
        except:
            return False

    def __len__(self):
        return len(self._node._findall(self._type))

    def __repr__(self):
        return repr(self._node._findall(self._type))


class BaseNode(anytree.Node):
    separator = ":"
    resolver = anytree.Resolver()

    class Type(enum.Enum):
        PARSER = 1
        REPOSITORY = 2
        MODULE = 3
        OPTION = 4

    def __init__(self, name, node_type, repository=None):
        anytree.Node.__init__(self, name)
        self._type = node_type
        self._functions = {}

        self._fullname = name
        self._filename = None

        # Dependency management
        self._repository = repository
        self._dependency_module_names = []
        self._dependencies_resolved = False
        self._dependencies = []

        self._description = ""
        # All _update()-able traits: defaults
        self._available_default = True
        self._format_description_default = format_description
        self._format_short_description_default = format_short_description
        self._context_default = None

        # All _update()-able traits: defaults
        self._available = (self._type != BaseNode.Type.MODULE)
        self._format_description = self._format_description_default
        self._format_short_description = self._format_short_description_default
        self._context = self._context_default
        self._ignore_patterns = lbuild.utils.default_ignore_patterns
        self._filters = lbuild.filter.default_filters

    @property
    def format_description(self):
        return format_description

    @property
    def format_short_description(self):
        return format_short_description

    @property
    def _filepath(self):
        return os.path.dirname(self._filename)

    @property
    def fullname(self):
        if self._fullname is None:
            self._fullname = self.name
        return self._fullname

    @property
    def type(self):
        return self._type

    @property
    def options(self):
        return self.all_options(depth=2)

    @property
    def submodules(self):
        return self.all_modules(depth=2)

    @property
    def repository(self):
        return self._repository

    @property
    def dependencies(self):
        if not self._dependencies_resolved:
            self._resolve_dependencies()
        return self._dependencies + [d for o in self.all_options(depth=2) for d in o._dependencies]

    @property
    def description(self):
        return self._format_description(self, str(self._description))

    @property
    def short_description(self):
        return self._format_short_description(self, str(self._description))

    @description.setter
    def description(self, description):
        self._description = description

    @property
    def option_value_resolver(self):
        resolver = NameResolver(self, self.Type.OPTION)
        resolver._value_resolver = True
        return resolver

    @property
    def option_resolver(self):
        return NameResolver(self, self.Type.OPTION)

    @property
    def module_resolver(self):
        return NameResolver(self, self.Type.MODULE)

    def render(self):
        return Renderer(self)

    def add_dependencies(self, *dependencies):
        """
        Add a new dependencies.

        The module name has not to be fully qualified.
        """
        self._dependency_module_names += dependencies

    def add_option(self, option):
        """
        Define new option for this module.

        The module options only influence the build process but not the
        selection and dependencies of modules.
        """
        if option.name in [c.name for c in self.children]:
            raise LbuildException("Option name '{}' is already defined".format(option.name))
        option._repository = self._repository
        option.parent = self
        option.add_dependencies(self.fullname)
        option._fullname = self.fullname + ":" + option.name

    def all_options(self, depth=None):
        return self._findall(self.Type.OPTION, depth)

    def all_modules(self, depth=None):
        return self._findall(self.Type.MODULE, depth)

    def _findall(self, node_type, depth=None):
        def _filter(n):
            return (n._type == node_type and
                    n._available and
                    n is not self)
        return anytree.search.findall(self, maxlevel=depth, filter_=_filter)

    def _resolve_dependencies(self, ignore_failure=False):
        """
        Update the internal list of dependencies.

        Resolves the module names to the actual module objects.
        """
        if self._dependencies_resolved:
            return
        dependencies = set()
        # print(self.fullname, self._dependency_module_names)
        dependency_names = set(n for n in self._dependency_module_names if ":" in n)
        for dependency_name in dependency_names:
            try:
                dependencies.add(self.module_resolver[dependency_name])
            except LbuildException as b:
                if not ignore_failure:
                    raise LbuildException("Cannot resolve dependencies!\n" + str(b))
                print("ignoring", dependency_name)
        self._dependencies = list(dependencies)
        self._dependencies_resolved = not ignore_failure
        for child in self.children:
            child._resolve_dependencies(ignore_failure)

    def _resolve_partial_max(self, query, max_results=1):
        nodes = self._resolve_partial(query, None)
        if nodes is None:
            raise LbuildException("Unknown '{}' in module '{}'!".format(query, self.fullname))
        if len(nodes) > max_results:
            raise LbuildException("Ambiguous '{}'! Found: '{}'".format(query, "', '".join([n.fullname for n in nodes])))
        return nodes

    def _resolve_partial(self, query, default=[]):
        # Try if query result is unique
        resolved1 = self._resolve(query, [])
        if len(resolved1) == 1:
            return resolved1
        # no result or ambiguous? try to fill the partial name
        query = ":".join(self._fill_partial_name(["" if p == "*" else p for p in query.split(":")]))
        resolved2 = self._resolve(query, [])
        if not (len(resolved2) or len(resolved1)):
            return default # neither found anything
        if not len(resolved2):
            return resolved1
        if not len(resolved1):
            return resolved2
        # return the less ambiguous one
        return resolved2 if len(resolved2) < len(resolved1) else resolved1

    def _resolve(self, query, default=[]):
        # :*   -> non-recursive
        # :**  -> recursive
        query = ":".join(p if len(p) else "*" for p in query.split(":"))
        try:
            qquery = ":" + query.replace(":**", "")
            if self.root._type == self.Type.PARSER:
                qquery = ":lbuild" + qquery
            # print("\n\n\n", qquery)
            found_modules = BaseNode.resolver.glob(self.root, qquery)
        except (anytree.resolver.ChildResolverError, anytree.resolver.ResolverError):
            return default
        modules = found_modules
        if query.endswith(":**"):
            for module in found_modules:
                modules.extend(module.descendants)
        # print("\n\n\n", modules)
        return modules if len(modules) else default

    def _fill_partial_name(self, partial_name):
        """
        Fill the array of the module name with the parts of the full name
        of the current module.

        Returns an array of the full name.
        """
        module_fullname_parts = self.fullname.split(":")

        # if partial_name is just leaf name, set scope to local node
        if len(partial_name) == 1:
            partial_name = module_fullname_parts + partial_name
        # Limit length of the module name to the length of the requested name
        depth = len(partial_name)
        if len(module_fullname_parts) > depth:
            module_fullname_parts = module_fullname_parts[:depth]

        # Using zip_longest restricts the name to the length of full name
        # if it is shorted than the requested module name.
        name = []
        for part, fill in itertools.zip_longest(partial_name,
                                                module_fullname_parts,
                                                fillvalue=""):
            name.append(fill if (part == "") else part)
        return name

    def _update_attribute(self, attr):
        self_attr = getattr(self, attr, "unknown")
        parent_attr = getattr(self.parent, attr, "unknown")
        if self_attr is "unknown" or parent_attr is "unknown":
            raise LbuildException("Cannot update non-existant attribute '{}'!".format(attr))

        if isinstance(self_attr, list):
            self_attr = list(set(self_attr + parent_attr))
            return
        if isinstance(self_attr, dict):
            self_attr.update(parent_attr)
            return

        default = getattr(self, attr + "_default")
        if ((parent_attr is not default) and (self_attr is default)):
            setattr(self, attr, parent_attr)
            # print("Updating {}.{} = {} -> {}.".format(self.fullname, attr, self_attr, parent_attr))

    def _update_format(self):
        if self.parent:
            self._update_attribute("_format_description")
            self._update_attribute("_format_short_description")
        for c in self.children:
            c._update_format()

    def _update(self):
        if self.parent:
            self._update_attribute("_format_description")
            self._update_attribute("_format_short_description")
            self._update_attribute("_available")
            self._update_attribute("_ignore_patterns")
            self._update_attribute("_filters")
            self._update_attribute("_context")
        for c in self.children:
            c._update()

    def _relocate_relative_path(self, path):
        """
        Relocate relative paths to the path of the repository
        configuration file.
        """
        if not os.path.isabs(path):
            path = os.path.join(self._filepath, path)
        return os.path.normpath(path)



#!/usr/bin/env python3
#
# Copyright (c) 2016-2018, Fabian Greif
# Copyright (c) 2018, Niklas Hauser
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

import enum
import inspect
import textwrap
import collections

import lbuild.filter
from .exception import LbuildException
from .node import BaseNode
from .format import color_str as c


class Option(BaseNode):
    def __init__(self, name, description, default=None, dependencies=None, convert_input=str, convert_output=str):
        BaseNode.__init__(self, name, BaseNode.Type.OPTION)
        self._dependency_handlers = lbuild.utils.listify(dependencies)
        self._description = description
        self._in = convert_input
        self._out = convert_output
        self._input = None
        self._output = None
        self._default = None
        self._set_default(default)

    def _set_default(self, default):
        if default is not None:
            self._input = self._in(default)
            self._output = self._out(default)
            self._default = self._in(default)

    def _update_dependencies(self):
        self._dependencies_resolved = False
        for handler in self._dependency_handlers:
            self._dependency_module_names += lbuild.utils.listify(handler(self._input))

    def _set_value(self, value):
        self._input = self._in(value)
        self._output = self._out(value)
        self._update_dependencies()

    @property
    def module(self):
        return self.parent

    @property
    def value(self):
        return self._output

    @value.setter
    def value(self, value):
        self._set_value(value)

    @property
    def values(self):
        return "String"

    def is_default(self):
        return self._input == self._default

    def format_value(self):
        value = str(self._input).strip(" \n")
        if value is "": value = '""';
        return value

    def format_values(self):
        if self.is_default() or self._default == "":
            return c("String")
        return c("String: ") + c(str(self._default)).wrap("underlined")

    def __lt__(self, other):
        return self.fullname.__lt__(other.fullname)

    def __str__(self):
        return self.fullname

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.fullname)


class StringOption(Option):
    def __init__(self, name, description, default=None, dependencies=None):
        Option.__init__(self, name, description, default, dependencies)


class BooleanOption(Option):
    def __init__(self, name, description, default=False, dependencies=None):
        Option.__init__(self, name, description, default, dependencies,
                        convert_input=self.as_boolean,
                        convert_output=self.as_boolean)

    @property
    def values(self):
        return [True, False]

    def format_values(self):
        if self._default:
            return c("True").wrap("underlined") + c(", False")
        return c("True, ") + c("False").wrap("underlined")

    @staticmethod
    def as_boolean(value):
        if value is None:
            return value
        elif isinstance(value, bool):
            return value
        elif str(value).lower() in ['true', 'yes', '1']:
            return True
        elif str(value).lower() in ['false', 'no', '0']:
            return False

        raise LbuildException("Value '{}' ({}) must be boolean!".format(value, type(value).__name__))


class NumericOption(Option):
    def __init__(self, name, description, minimum=None, maximum=None, default=None, dependencies=None):
        Option.__init__(self, name, description, default, dependencies,
                        convert_input=str,
                        convert_output=self.as_numeric_value)
        self.minimum = minimum
        self.maximum = maximum
        if self.minimum is not None and self.maximum is not None:
            if self.minimum >= self.maximum:
                raise LbuildException("Minimum '{}' must be smaller than maximum '{}'!".format(self.minimum, self.maximum))

    @Option.value.setter
    def value(self, value):
        numeric_value = self.as_numeric_value(value)
        if self.minimum is not None and numeric_value < self.minimum:
            raise LbuildException("Value '{}' of '{}' must be greater than '{}'".format(
                                  numeric_value, self.fullname, self.minimum))
        if self.maximum is not None and numeric_value > self.maximum:
            raise LbuildException("Value '{}' of '{}' must be smaller than '{}'".format(
                                  numeric_value, self.fullname, self.maximum))
        self._set_value(value)

    @property
    def values(self):
        return [self.minimum, self.maximum]

    def format_values(self):
        minimum = c("-Inf" if self.minimum is None else str(self.minimum))
        maximum = c("+Inf" if self.maximum is None else str(self.maximum))
        if self._default is None:
            return minimum + c(" ... ") + maximum

        default = c(str(self._default)).wrap("underlined")
        if minimum != default and maximum != default:
            return minimum + c(" .. ") + default + c(" .. ") + maximum

        if maximum == default:
            return minimum + c(" ... ") + default

        if minimum == default:
            return default + c(" ... ") + maximum

        return default

    @staticmethod
    def as_numeric_value(value):
        if value is None:
            return value
        elif isinstance(value, (int, float)):
            return value
        elif isinstance(value, str):
            try:
                return int(value, 0)
            except ValueError:
                pass

        raise LbuildException("Value '{}' ({}) must be numeric!".format(value, type(value).__name__))


class EnumerationOption(Option):
    def __init__(self, name, description, enumeration, default=None, dependencies=None):
        Option.__init__(self, name, description, None, dependencies,
                        convert_input=self._obj_to_str,
                        convert_output=self.as_enumeration)
        if self._is_enum(enumeration):
            self._enumeration = {self._obj_to_str(entry): entry.value for entry in enumeration}
        elif isinstance(enumeration, (list, tuple, set, range)) and \
           len(enumeration) == len(set(enumeration)):
            # If the argument is a list and the items in the list are unique,
            # convert it so that the value of the enum equals its name.
            self._enumeration = {self._obj_to_str(entry): entry for entry in enumeration}
        elif isinstance(enumeration, dict):
            self._enumeration = enumeration
            for key in self._enumeration:
                if not isinstance(key, str):
                    raise LbuildException("All enumeration keys must be of type string!")
        else:
            raise LbuildException("Type {} currently not supported".format(type(enumeration).__name__))

        self._set_default(default)

    @staticmethod
    def _is_enum(obj):
        return inspect.isclass(obj) and issubclass(obj, enum.Enum)

    @staticmethod
    def _obj_to_str(obj):
        if EnumerationOption._is_enum(obj):
            return str(obj.name)
        return str(obj).strip()

    def _format_values(self):
        values = list(map(str, self._enumeration.keys()))
        values.sort(key=lambda v: (float(v) if v.isdigit() else 0, v))
        return values

    def format_values(self):
        values = [c(v).wrap("underlined") if v == self._default else c(v) for v in self._format_values()]
        return c(", ").join(values)

    def as_enumeration(self, value):
        try:
            return self._enumeration[self._obj_to_str(value)]
        except KeyError:
            raise LbuildException("Value '{}' not found in enumeration '{}'. " \
                                  "Possible values are:\n'{}'.".format(self.fullname, value, "', '".join(self._enumeration)))


class SetOption(EnumerationOption):
    def __init__(self, name, description, enumeration, default=None, dependencies=None):
        EnumerationOption.__init__(self, name, description, enumeration, None, dependencies)
        self._in = self.str_to_set
        self._out = self.as_set
        self._set_default(default)

    def format_value(self):
        return "{{{}}}".format(", ".join(map(str, self._output)))

    def format_values(self):
        values = [c(v).wrap("underlined") if v in self._default else c(v) for v in self._format_values()]
        return c(", ").join(values)

    @staticmethod
    def str_to_set(values):
        if isinstance(values, str):
            values = [v.strip() for v in values.split(",")]
        return list(map(str, lbuild.utils.listify(values)))

    def as_set(self, values):
        values = self.str_to_set(values)
        # remove duplicates, but retain order with OrderedDict
        return [self.as_enumeration(value) for value in collections.OrderedDict.fromkeys(values)]




#!/usr/bin/env python3
#
# Copyright (c) 2015-2018, Fabian Greif
# Copyright (c) 2018, Niklas Hauser
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

import os
import logging
import itertools

import lbuild.utils
import lbuild.filter
import lbuild.option
import lbuild.repository

from . import exception

from .repository import Repository

from .exception import LbuildException
from .exception import LbuildForwardException

from .node import BaseNode
from .facade import ModuleInitFacade, ModulePrepareFacade

LOGGER = logging.getLogger('lbuild.module')


class ModuleBase:
    pass


def load_module_from_file(repository, repo_options, filename, parent=None):
    module = ModuleInit(repository, filename)
    if parent: module.parent = parent;
    module.functions = lbuild.node.load_functions_from_file(
        repository, filename,
        required=['init', 'prepare', 'build'],
        optional=['pre_build', 'validate', 'post_build'],
        local={
            'Module': ModuleBase,
            'PreBuildException': lbuild.exception.LbuildValidateException,
            'ValidateException': lbuild.exception.LbuildValidateException,
        })
    module.init()
    return module.prepare(repo_options)


def load_module_from_object(repository, repo_options, module_obj, filename=None, parent=None):
    module = ModuleInit(repository, filename)
    if parent: module.parent = parent;
    module.functions = lbuild.utils.get_global_functions(module_obj,
            required=['init', 'prepare', 'build'],
            optional=['pre_build', 'validate', 'post_build'])
    module.init()
    return module.prepare(repo_options)


def build_modules(initmodules):
    rmodules = {}
    modules = []
    # First convert the modules into node objects
    for initmodule in initmodules:
        module = Module(initmodule)
        rmodules[module.fullname] = module
        modules.append(module)

    # then connect the entire tree
    for module in modules:
        parent = ":".join(module.fullname.split(":")[:-1])
        module.parent = rmodules[parent] if ":" in parent else module._repository

    # Now update the tree
    for module in modules:
        module._update()

    return modules


class ModuleInit:
    def __init__(self, repository, filename=None):
        self.filename = os.path.realpath(filename)
        self.filepath = os.path.dirname(self.filename) if filename else None
        self.repository = repository

        self.name = None
        self.parent = self.repository.name
        self.description = ""
        self.functions = {}
        self.available = False

        self._submodules = []
        self._options = []
        self._dependencies = []
        self._filters = {}

    @property
    def fullname(self):
        return self.parent + ":" + self.name

    def init(self):
        # Execute init() function from module to get module name
        lbuild.utils.with_forward_exception(self, lambda: self.functions['init'](ModuleInitFacade(self)))

        if self.name is None:
            raise LbuildException("The init(module) function must set a module name! " \
                                "Please set the 'name' attribute.")

        if self.parent.startswith(":"):
            self.parent = self.repository.name + self.parent
        if not self.parent.startswith(self.repository.name):
            self.parent = self.repository.name + ":" + self.parent

    def prepare(self, repo_options):
        is_available = lbuild.utils.with_forward_exception(self,
                lambda: self.functions["prepare"](ModulePrepareFacade(self), self.repository.option_value_resolver))

        available_modules = []
        if is_available is None:
            raise LbuildException("The prepare() function for module '{}' must "
                                "return True or False."
                                .format(self.name))
        elif is_available:
            available_modules.append(self)
        self.available = is_available

        for submodule in self._submodules:
            if isinstance(submodule, ModuleBase):
                modules = load_module_from_object(repository=self.repository,
                                                  repo_options=repo_options,
                                                  module_obj=submodule,
                                                  filename=self.filename,
                                                  parent=self.fullname)
            else:
                modules = load_module_from_file(repository=self.repository,
                                                repo_options=repo_options,
                                                filename=os.path.join(self.filepath, submodule),
                                                parent=self.fullname)

            available_modules.extend(modules)

        return available_modules


class Module(BaseNode):
    def __init__(self, module: ModuleInit):
        """
        Create new module definition.

        Args:
            repository: Parent repository of the module.
            filename: Full path of the module file.
            path: Path to the module file. Used as base for relative paths
                during the building step of the module.
        """
        BaseNode.__init__(self, module.name, self.Type.MODULE, module.repository)
        self._filename = module.filename
        self._functions = module.functions
        self._description = module.description
        self._fullname = module.fullname
        self._available = module.available
        self._filters.update({self._repository.name + ":" + name: func for name, func in module._filters})

        # OptionNameResolver defined in the module configuration file. These
        # options are configurable through the project configuration file.
        for o in module._options:
            self.add_option(o)

        self.add_dependencies(*module._dependencies)
        if ":" in module.parent:
            self.add_dependencies(module.parent)

    def validate(self, env):
        validate = self._functions.get("validate", self._functions.get("pre_build", None))
        if validate is not None:
            LOGGER.info("Validate " + self.fullname)
            lbuild.utils.with_forward_exception(self, lambda: validate(env))

    def build(self, env):
        LOGGER.info("Build " + self.fullname)
        lbuild.utils.with_forward_exception(self, lambda: self._functions["build"](env))

    def post_build(self, env, buildlog):
        post_build = self._functions.get("post_build", None)
        if post_build is not None:
            LOGGER.info("Post-Build " + self.fullname)
            lbuild.utils.with_forward_exception(self, lambda: post_build(env, buildlog))

    def __lt__(self, other):
        return self.fullname.__lt__(other.fullname)

    def __repr__(self):
        return "Module({})".format(self.fullname)

    def __str__(self):
        return self.fullname



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015-2018, Fabian Greif
# Copyright (c) 2018, Niklas Hauser
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

import os
import glob
import shutil
import logging
import fnmatch

import lbuild.filter
import lbuild.utils

from .facade import RepositoryInitFacade, RepositoryPrepareFacade
from .node import BaseNode, NameResolver
from .exception import LbuildException
from . import utils

LOGGER = logging.getLogger('lbuild.repository')


class Repository(BaseNode):
    """
    A repository is a set of modules.
    """

    def __init__(self, filename, name=None):
        """
        Construct a new repository object.

        At the construction time of the object, the name of repository may not
        be known e.g. if the repository is loaded from a `repo.lb` file.
        """
        BaseNode.__init__(self, name, self.Type.REPOSITORY, self)
        # Path to the repository file. All relative paths refer to this path.
        self._filename = os.path.realpath(filename)
        self._config_map = {}
        self._new_filters = {}

        # List of module filenames which are later transfered into
        # module objects
        self._module_files = []

    @property
    def modules(self):
        return {m.fullname:m for m in self.all_modules()}

    @staticmethod
    def parse_repository(repofilename: str):
        LOGGER.debug("Parse repository '%s'", repofilename)

        repo = Repository(repofilename)
        repo._functions = lbuild.node.load_functions_from_file(
                repo, repofilename, required=['init', 'prepare'], optional=['build'])

        # Execution init() function. In this function options are added.
        repo._functions['init'](RepositoryInitFacade(repo))

        if repo.name is None:
            raise LbuildException("The init(repo) function must set a repository name! "
                                "Please write the 'name' attribute.")

        # Prefix the global filters with the `repo.` name
        for name, func in repo._new_filters.items():
            repo._filters[repo.name + "." + name] = func

        # Prefix the global configuration map with the `repo:` name
        config_map = repo._config_map.copy()
        repo._config_map = {}
        for name, path in config_map.items():
            repo._config_map[repo.name + ":" + name] = path

        return repo

    def prepare(self, options):
        lbuild.utils.with_forward_exception(self,
                lambda: self._functions["prepare"](RepositoryPrepareFacade(self), self.option_value_resolver))

        modules = []
        # Parse the modules inside this repository
        for modulefile in self._module_files:
            module = lbuild.module.load_module_from_file(self, options, modulefile)
            modules.extend(module)

        return modules

    def build(self, env):
        build = self._functions.get("build", None)
        if build is not None:
            lbuild.utils.with_forward_exception(self, lambda: build(env))

    def add_modules_recursive(self, basepath="", modulefile="module.lb", ignore=None):
        """
        Find all module files following a specific pattern.

        Args:
            basepath: Rootpath for the search.
            modulefile: Filename pattern of the module files to search
                for (default: "module.lb").
            ignore: Filename pattern to ignore during search
        """
        ignore = utils.listify(ignore) + self._ignore_patterns
        ignore.append(os.path.relpath(self._filename, self._filepath))
        basepath = self._relocate_relative_path(basepath)
        for path, _, files in os.walk(basepath):
            for file in files:
                if any(fnmatch.fnmatch(file, i) for i in ignore):
                    continue
                if fnmatch.fnmatch(file, modulefile):
                    modulefilepath = os.path.normpath(os.path.join(path, file))
                    self._module_files.append(modulefilepath)

    def add_modules(self, *modules):
        """
        Add one or more module files.

        Args:
            modules: List of filenames
        """
        modules = [lbuild.utils.listify(m) for m in modules]
        modules = [inner for outer in modules for inner in outer]
        for file in modules:
            file = self._relocate_relative_path(file)

            if not os.path.isfile(file):
                raise LbuildException("Module file not found '%s'" % file)

            self._module_files.append(file)

    def glob(self, pattern):
        pattern = os.path.abspath(self._relocate_relative_path(pattern))
        return glob.glob(pattern)

    def __lt__(self, other):
        return self.fullname.__lt__(other.fullname)

    def __repr__(self):
        return "Repository({})".format(self._filepath)

    def __str__(self):
        """ Get string representation of repository.  """
        if self.name is not None:
            # The name must not always be set, e.g. during load
            # the repository name is only known after calling the
            # init function.
            return "{} at {}".format(self.name, self._filepath)
        else:
            return self._filepath



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015-2018, Fabian Greif
# Copyright (c) 2018, Niklas Hauser
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

import os
import anytree
import pkgutil
import logging
import collections
import lxml.etree
from os.path import realpath, join
from pathlib import Path

from .exception import LbuildException

import lbuild.module

LOGGER = logging.getLogger('lbuild.config')
DEFAULT_CACHE_FOLDER = ".lbuild_cache"

class ConfigNode(anytree.AnyNode):
    def __init__(self, parent=None):
        anytree.AnyNode.__init__(self, parent, filename=Path())
        self._cachefolder = Path()
        self._extends = collections.defaultdict(list)
        self._vcs = []
        self._repositories = []
        self._modules = []
        self._options = {}

    @property
    def repositories(self):
        return self._repositories

    @property
    def modules(self):
        return self._modules

    @property
    def options(self):
        return self._options

    @property
    def vcs(self):
        return self._vcs

    @property
    def cachefolder(self):
        return self._cachefolder

    def add_commandline_options(self, cmd_options):
        for option in cmd_options:
            parts = option.split('=')
            self._options[parts[0]] = parts[1]

    def _flatten(self, config):
        for node in (list(self.siblings) + [self]):
            config._cachefolder = node._cachefolder
            config._extends.update(node._extends)
            config._vcs.extend(node._vcs)
            config._repositories.extend(node._repositories)
            config._modules.extend(node._modules)
            config._options.update(node._options)
        if self.parent:
            self.parent._flatten(config)

    def flatten(self):
        config = ConfigNode()
        self.last._flatten(config)
        config._repositories = list(set(config._repositories))
        config._modules = list(set(config._modules))
        return config

    def extend(self, node, config):
        if config and node:
            below = node.children[0] if len(node.children) else None
            config.parent = node
            if below:
                below.parent = config

    def extend_last(self, config):
        self.extend(self.last, config)

    @property
    def last(self):
        descendants = self.root.descendants
        return descendants[-1] if len(descendants) else self

    def find(self, filename):
        return anytree.find_by_attr(self.root, name="filename", value=filename)

    def render(self):
        if self.filename == Path():
            return "ConfigNode(Empty)"
        return anytree.RenderTree(self, anytree.ContRoundStyle())

    @staticmethod
    def from_filesystem(startpath=None, name="lbuild.xml"):
        startpath = Path(startpath) if startpath else Path.cwd()
        while startpath.exists():
            config = (startpath / name)
            if config.exists():
                return ConfigNode.from_file(config)
            if startpath.parent == startpath:
                break
            startpath = startpath.parent
        return None

    @staticmethod
    def from_file(configfile, parent=None, fail_silent=False):
        filename = os.path.relpath(str(configfile), os.getcwd())
        LOGGER.debug("Parse configuration '{}'".format(filename))
        if fail_silent and not os.path.exists(filename):
            return ConfigNode()

        xmltree = ConfigNode._load_and_verify(configfile)

        config = ConfigNode(parent)
        config.filename = filename
        configpath = str(Path(config.filename).parent)
        # load extend strings
        for e in xmltree.iterfind("extends"):
            cpath = Path(ConfigNode._rel_path(e.text, configpath))
            if cpath.exists():
                ConfigNode.from_file(str(cpath), config)
            else:
                config._extends[config.filename].append(e.text)

        # Load cachefolder
        cache_node = xmltree.find("repositories/cache")
        if cache_node:
            config._cachefolder = ConfigNode._rel_path(cache_node.text, configpath)
        else:
            config._cachefolder = join(configpath, DEFAULT_CACHE_FOLDER)

        # Load version control nodes
        for vcs_node in xmltree.iterfind("repositories/repository/vcs"):
            config._vcs += [ConfigNode.to_dict(vcs) for vcs in vcs_node.iterchildren()]

        # Load repositories
        for path_node in xmltree.iterfind("repositories/repository/path"):
            repopath = path_node.text.format(cache=config._cachefolder)
            filename = realpath(join(configpath, repopath))
            config._repositories.append(filename)

        # Load all requested modules
        config._modules = xmltree.xpath('modules/module/text()')

        # Load options
        for option_node in xmltree.xpath('options/option'):
            name = option_node.attrib['name']
            value = option_node.attrib.get('value', option_node.text)
            config._options[name] = value

        return config

    @staticmethod
    def _load_and_verify(configfile):
        try:
            xmlroot = lxml.etree.parse(str(configfile))

            xmlschema = lxml.etree.fromstring(pkgutil.get_data('lbuild', 'resources/configuration.xsd'))
            schema = lxml.etree.XMLSchema(xmlschema)
            schema.assertValid(xmlroot)

            xmltree = xmlroot.getroot()
        except OSError as error:
            raise LbuildException(error)
        except (lxml.etree.DocumentInvalid,
                lxml.etree.XMLSyntaxError,
                lxml.etree.XMLSchemaParseError,
                lxml.etree.XIncludeError) as error:
            raise LbuildException("While parsing '{}': {}".format(
                                  error.error_log.last_error.filename, error))
        return xmltree

    @staticmethod
    def _rel_path(path, configpath):
        if os.path.isabs(path):
            return path
        else:
            return join(configpath, path)

    @staticmethod
    def to_dict(xmltree):
        """
        Convert XML to a Python dictionary according to
        http://www.xml.com/pub/a/2006/05/31/converting-between-xml-and-json.html
        """
        d = {xmltree.tag: {} if xmltree.attrib else None}
        children = []
        for c in xmltree:
            children.append(c)
        if children:
            dd = collections.defaultdict(list)
            for dc in [ConfigNode.to_dict(c) for c in  children]:
                for k, v in dc.items():
                    dd[k].append(v)
            d = {xmltree.tag: {k:v[0] if len(v) == 1 else v for k, v in dd.items()}}
        if xmltree.attrib:
            d[xmltree.tag].update(('@' + k, v) for k, v in xmltree.attrib.items())
        if xmltree.text:
            text = xmltree.text.strip()
            if children or xmltree.attrib:
                if text:
                    d[xmltree.tag]['#text'] = text
            else:
                d[xmltree.tag] = text
        return d




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2016, Fabian Greif
# Copyright (c) 2018, Niklas Hauser
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

import re
import collections

import lbuild.parser

def get_valid_identifier(name):
    return re.sub(r'\W|^(?=\d)', '_', name)


def graphviz(parser, selected_modules, depth, clustered=False):
    output = []
    output.append("digraph dependencies")
    output.append("{")
    # output.append("\trankdir=LR;")

    modules = parser.resolve_dependencies(selected_modules, depth)

    # Check whether all available modules are selected for displaying
    all_selected = (len(parser.modules) == len(modules))

    # Sort modules by repository
    repositories = collections.defaultdict(list)
    for module in modules:
        repositories[module.repository.name].append(module)

    for repo_name, modules in repositories.items():
        if clustered:
            output.append("\tsubgraph cluster{}".format(get_valid_identifier(repo_name)))
        else:
            output.append("\tsubgraph {}".format(get_valid_identifier(repo_name)))

        output.append("\t{")
        output.append("\t\tlabel = \"{}\";".format(repo_name))
        output.append("\t\tnode [style=filled, shape=box];")
        output.append("")
        for module in sorted(modules):
            if clustered:
                # Remove the repository name for the clusted output. The
                # repository name is meantioned in the cluster already.
                name = ":\\n".join(module.fullname.split(":")[1:])
            else:
                name = ":\\n".join(module.fullname.split(":"))

            attributes = []
            attributes.append("label=\"{}\"".format(name))

            if not all_selected:
                if module in selected_modules:
                    attributes.append('style="filled,bold"')
                else:
                    for dep in sorted(module.dependencies):
                        if dep not in modules:
                            attributes.append('style="filled,dashed"')
                            break
                    else:
                        attributes.append('style="filled,solid"')

            output.append("\t\t{} [{}];".format(get_valid_identifier(module.fullname),
                                                ", ".join(attributes)))
        output.append("\t}")

    for module in sorted(modules):
        for dep in sorted(module.dependencies):
            if not all_selected:
                if dep not in modules:
                    # Dependencies may not be drawn if the depth of the
                    # dependency resolution is too small.
                    continue

            output.append("\t{} -> {};".format(get_valid_identifier(module.fullname),
                                               get_valid_identifier(dep.fullname)))

    output.append("}")
    return "\n".join(output)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2016, Fabian Greif
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

import logging.config

def configure_logger(verbosity):
    """
    Load the default configuration for the logger.
    """
    logging.config.dictConfig({
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'full': {
                # 'format': '%(asctime)s [%(levelname)s] %(name)s: %(message)s'
                'format': '[%(levelname)s] %(name)s: %(message)s'
            },
            'simple': {
                'format': '%(message)s'
            },
        },
        'handlers': {
            'default': {
                'class':'logging.StreamHandler',
                'formatter': 'full',
            },
        },
        'loggers': {
            '': {
                'handlers': ['default'],
                'level': 'DEBUG' if verbosity > 1 else ('INFO' if verbosity == 1 else 'WARNING'),
                'propagate': True
            }
        }
    })



#!/usr/bin/env python3
#
# Copyright (c) 2015, Fabian Greif
# Copyright (c) 2018, Niklas Hauser
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

from setuptools import setup, find_packages
from lbuild.__init__ import __version__

with open("README.md") as f:
    long_description = f.read()

setup(
    name = "lbuild",
    version = __version__,
    python_requires=">=3.5.0",
    entry_points={
        "console_scripts": [
            "lbuild = lbuild.main:main",
        ],
    },

    packages = find_packages(exclude=["test"]),

    include_package_data = True,

	# Make sure all files are unzipped during installation
	#zip_safe = False,

    install_requires = ["lxml", "jinja2", "gitpython", "anytree", "colorful"],

    extras_require = {
        "test": ["testfixtures", "coverage"],
    },

    # Metadata
    author = "Fabian Greif, Niklas Hauser",
    author_email = "fabian.greif@rwth-aachen.de, niklas@salkinium.com",
    description = "Generic, modular code generator using the Jinja2 template engine.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license = "BSD",
    keywords = "library builder generator",
    url = "https://github.com/modm-io/lbuild",
    classifiers = [
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Software Development",
        "Topic :: Software Development :: Code Generators",
        "Topic :: Software Development :: Embedded Systems",
    ],
)



#!/usr/bin/env python3
#
# Copyright (c) 2015-2017, Fabian Greif
# Copyright (c) 2018, Niklas Hauser
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.


class LbuildException(Exception):
    """
    Base class for exception thrown by lbuild.
    """
    pass


class LbuildArgumentException(LbuildException):
    """
    Raised if something is wrong with the command line arguments.
    """
    pass


class LbuildOptionFormatException(LbuildException):
    """
    Exception for all invalid option names.
    """
    def __init__(self, name):
        LbuildException.__init__(self,
                               "Invalid option format for '{}'. Option must "
                               "options are one (repository option) or "
                               "two and more (module option) colons.".format(name))


class LbuildAttributeException(LbuildException):
    def __init__(self, name):
        LbuildException.__init__(self,
                               ("The attribute '{}' may only be changed in "
                                "the init(...) method".format(name)))


class LbuildTemplateException(LbuildException):
    """
    Error in Jinja2 template evaluation.
    """
    pass


class LbuildValidateException(LbuildException):
    pass


class LbuildBuildException(LbuildException):
    """
    Exceptions raised during the build of project.

    E.g. when a previously generated file is being overwritten by another
    module.
    """
    pass


class LbuildAggregateException(LbuildException):
    """
    Collection of multiple exceptions.
    """
    def __init__(self, exceptions):
        self.exceptions = exceptions


class LbuildForwardException(LbuildException):
    """
    Handler for regular Python exception thrown in user defined functions.

    Use to forward the exception to the command line interface and provide
    additional informations.
    """
    def __init__(self, module, exception):
        self.module = module
        self.exception = exception



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015, 2017-2018, Fabian Greif
# Copyright (c) 2018, Niklas Hauser
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

import os
import re
import sys
import uuid
import shutil
import fnmatch
import colorful
import importlib.util
import importlib.machinery

from .exception import LbuildException
from .exception import LbuildForwardException

import lbuild

default_ignore_patterns = [
    "*/.git*",
    "*/.DS_Store*",
    "*/__pycache__",
    "*/module.lb",
    "*/repo.lb",
    "*/*.pyc"
]

def ignore_files(*files):
    """
    Ignore file and folder names without checking the full path.

    Example: the following code with ignore all files with the ending `.lb`:
    ```
    env.copy(".", ignore=env.ignore_files("*.lb"))
    ```

    Based on the shutil.ignore_patterns() function.
    """
    return shutil.ignore_patterns(*files)

def ignore_patterns(*patterns):
    """
    Ignore patterns based on the absolute file path.

    Use an `*` at the beginning to match relative paths:
    ```
    env.copy(".", ignore=env.ignore_patterns("*platform/*.lb"))
    ```
    This ignores all files in the `platform` sub-directory with the
    ending `.lb`.
    """

    def check(path, files):
        ignored = set()
        for pattern in patterns:
            for filename in files:
                if fnmatch.fnmatch(os.path.join(path, filename), pattern):
                    # The copytree function uses only the filename to check
                    # which files should be ignored, not the absolute path.
                    ignored.add(filename)
        return ignored

    return check


def _listify(obj):
    if obj is None:
        return list()
    if isinstance(obj, (list, tuple, set, range)):
        return list(obj)
    if hasattr(obj, "__iter__") and not hasattr(obj, "__getitem__"):
        return list(obj)
    return [obj, ]


def listify(*objs):
    return [l for o in objs for l in _listify(o)]


def get_global_functions(env, required, optional=None):
    """
    Get global functions an environment.

    Args:
        required: List of required functions.
        optional: List of optional functions.
    """
    def get(name, required=True):
        if isinstance(env, dict):
            val = env.get(name, None)
        else:
            val = getattr(env, name, None)
        if required and val is None:
            raise LbuildException("No function '{}' found!".format(functionname))
        return val

    if optional is None: optional = [];
    functions = {}
    for name in (required + optional):
        function = get(name, name in required)
        if function is not None:
            functions[name] = function

    return functions


def load_module_from_file(filename, local, modulename=None):
    """
    Load a python module from a local file.

    Args:
        filename: Name of the file to load. The file extension is ignored.
        local: dictionary of symbols which will be added to the global
            namespace when executing the module code.
        modulename: Name of the module. When set to `None`.

    Returns:
        Namespace of the module.
    """
    # The actual name of the module is only known after it is loaded. Therefore
    # a UUID is used instead here.
    if modulename is None:
        modulename = "lbuild.modules.{}".format(uuid.uuid1())

    loader = importlib.machinery.SourceFileLoader(modulename, filename)

    spec = importlib.util.spec_from_loader(loader.name, loader)
    try:
        module = importlib.util.module_from_spec(spec)
    except AttributeError:
        import types
        module = types.ModuleType(modulename)

    # Prepare the environment of the module. Everything set here will
    # be available in the global namespace when executing the module.
    module.__dict__.update(local)

    # Load the module. This executes the code inside the lbuild module file.
    loader.exec_module(module)

    sys.modules[modulename] = module
    return module.__dict__


def with_forward_exception(module, function):
    """
    Run a function a store exceptions as forward exceptions.
    """
    try:
        return function()
    except LbuildException as error:
        raise
    except Exception as error:
        # Forward all exception which are not LbuildExceptions
        raise LbuildForwardException(module, error)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015-2018, Fabian Greif
# Copyright (c) 2018, Niklas Hauser
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

import os
import sys
import argparse
import textwrap
import colorful
import traceback

import lbuild.parser
import lbuild.logger
import lbuild.module
import lbuild.vcs.common
from lbuild.format import format_option_short_description

__version__ = '1.0.4'


class InitAction:
    config_required = True

    def register(self, argument_parser):
        parser = argument_parser.add_parser("init",
            help="Load remote repositories into the cache folder.")
        parser.set_defaults(execute_action=self.perform)

    def perform(self, args, config):
        lbuild.vcs.common.initialize(config)
        return ""


class UpdateAction:
    config_required = True

    def register(self, argument_parser):
        parser = argument_parser.add_parser("update",
            help="Update the content of remote repositories in the cache folder.")
        parser.set_defaults(execute_action=self.perform)

    def perform(self, args, config):
        lbuild.vcs.common.update(config)
        return ""


class ManipulationActionBase:
    """
    Base class for actions that interact directly with the parser repositories.
    """
    config_required = True

    def prepare_repositories(self, args, config):
        parser = lbuild.parser.Parser(config)
        parser.load_repositories(args.repositories)

        parser.config.add_commandline_options(args.options)
        repo_options = parser.merge_repository_options()

        return self.perform(args, parser, repo_options)


class DiscoverAction(ManipulationActionBase):
    config_required = False

    def register(self, argument_parser):
        parser = argument_parser.add_parser("discover",
            help="Render the available repository tree with modules and options. "
                 "You may need to provide options to see the entire tree!")
        parser.add_argument("-n", "--name",
            dest="names",
            type=str,
            action="append",
            default=[],
            help="Select a specific repository, module or option.")
        parser.set_defaults(execute_action=self.prepare_repositories)

    def perform(self, args, parser, repo_options):
        if not len(parser._undefined_options(repo_options)):
            parser.prepare_repositories(repo_options)
            parser.merge_module_options()

        if len(args.names):
            ostream = []
            for name in args.names:
                node = parser.find(name).description
                ostream.extend([node])
            return "\n\n\n\n".join(ostream)

        return parser.render()


class DiscoverOptionsAction(ManipulationActionBase):
    config_required = False

    def register(self, argument_parser):
        parser = argument_parser.add_parser("discover-options",
            help="Display all known option names, current values, allowed inputs and "
                 "short descriptions.")
        parser.add_argument("-n", "--name",
            dest="names",
            type=str,
            action="append",
            default=[],
            help="Select a specific repository or module.")
        parser.set_defaults(execute_action=self.prepare_repositories)

    def perform(self, args, parser, repo_options):
        if not len(parser._undefined_options(repo_options)):
            parser.prepare_repositories(repo_options)
            parser.merge_module_options()

        names = args.names if len(args.names) else ["*", ":**"]
        nodes = parser.find_any(names, (parser.Type.MODULE, parser.Type.REPOSITORY))
        options = [o for n in nodes for o in n.options]

        ostream = []
        for option in sorted(options, key=lambda n: (n.depth, n.fullname)):
            ostream.append(format_option_short_description(option))
            if option.short_description:
                ostream.append("")
                ostream.append(textwrap.indent(option.short_description, "  "))
                ostream.append("")

        return "\n".join(ostream)


def get_modules(parser, repo_options):
    modules = parser.prepare_repositories(repo_options)
    module_options = parser.merge_module_options()
    selected_modules = parser.find_modules(parser.config.modules)
    return parser.resolve_dependencies(selected_modules)


class ValidateAction(ManipulationActionBase):
    config_required = False

    def register(self, argument_parser):
        parser = argument_parser.add_parser("validate",
            help="Validate the library configuration and data inputs with the given options.")
        parser.add_argument("-m", "--module",
            dest="modules",
            type=str,
            action="append",
            default=[],
            help="Select a specific module.")
        parser.set_defaults(execute_action=self.prepare_repositories)

    def perform(self, args, parser, repo_options):
        parser.config.modules.extend(args.modules)
        build_modules = get_modules(parser, repo_options)
        parser.validate_modules(build_modules)
        return "Library configuration valid."


class BuildAction(ManipulationActionBase):
    config_required = False

    def register(self, argument_parser):
        parser = argument_parser.add_parser("build",
            help="Generate the library source code blob with the given options.")
        parser.add_argument("-m", "--module",
            dest="modules",
            type=str,
            action="append",
            default=[],
            help="Select a specific module.")
        parser.add_argument("--simulate",
            dest="simulate",
            action="store_true",
            default=False,
            help="Build, but do not write any files. Prints out all generated file names.")
        parser.add_argument("--no-log",
            dest="buildlog",
            action="store_false",
            default=True,
            help="Do not create a build log. This log contains all files being "
                 "generated, their source files and the module which generated "
                 "the file.")
        parser.set_defaults(execute_action=self.prepare_repositories)

    def perform(self, args, parser, repo_options):
        parser.config.modules.extend(args.modules)
        build_modules = get_modules(parser, repo_options)

        lbuild.environment.simulate = args.simulate
        buildlog = lbuild.buildlog.BuildLog(args.path)
        parser.build_modules(build_modules, buildlog)

        if args.simulate:
            ostream = []
            for op in buildlog.operations:
                ostream.append(op.local_filename_out())
            return "\n".join(sorted(ostream))
        elif args.buildlog:
            configfilename = args.config
            logfilename = configfilename + ".log"
            with open(logfilename, "wb") as logfile:
                logfile.write(buildlog.to_xml(to_string=True, path=os.getcwd()))

        return ""


class CleanAction(ManipulationActionBase):
    config_required = False

    def register(self, argument_parser):
        parser = argument_parser.add_parser("clean",
            help="Remove previously generated files.")
        parser.add_argument("--buildlog",
            dest="buildlog",
            help="Use the given buildlog to identify the files to remove.")
        parser.set_defaults(execute_action=self.prepare_repositories)

    def perform(self, args, parser, repo_options):
        ostream = []
        if args.buildlog is not None:
            with open(args.buildlog, "rb") as logfile:
                buildlog = lbuild.buildlog.BuildLog.from_xml(logfile.read(), path=os.getcwd())
        else:
            build_modules = get_modules(parser, repo_options)

            lbuild.environment.simulate = True
            buildlog = lbuild.buildlog.BuildLog(args.path)
            parser.build_modules(args.path, build_modules, buildlog)

        dirs = set()
        filenames = [op.local_filename_out() for op in buildlog.operations]
        for filename in sorted(filenames):
            ostream.append("Removing " + filename)
            dirs.add(os.path.dirname(filename))
            try:
                os.remove(filename)
            except Exception as e:
                pass

        dirs = sorted(list(dirs), key=lambda d: -d.count("/"))
        for di in dirs:
            try:
                os.removedirs(di)
            except Exception as e:
                pass

        return "\n".join(ostream)


class DependenciesAction(ManipulationActionBase):
    config_required = False

    def register(self, argument_parser):
        parser = argument_parser.add_parser("dependencies",
            help="Generate a grahpviz representation of the module dependencies.")
        parser.add_argument("-m", "--module",
            dest="modules",
            type=str,
            action="append",
            default=[],
            help="Select specific modules.")
        parser.add_argument("-n", "--depth",
            dest="depth",
            type=int,
            default=sys.maxsize,
            help="Only show dependencies up to a specific depth. Only valid if "
                 "specific modules are selected, otherwise all modules are printed "
                 "anyways.")
        parser.set_defaults(execute_action=self.prepare_repositories)

    def perform(self, args, parser, repo_options):
        available_modules = parser.prepare_repositories(repo_options)

        if len(args.modules) == 0:
            selected_modules = [":**"]
        else:
            selected_modules = args.modules
        selected_modules = parser.find_modules(selected_modules)
        dot_file = lbuild.builder.dependency.graphviz(parser,
                                                      selected_modules,
                                                      args.depth,
                                                      clustered=False)
        return dot_file


def prepare_argument_parser():
    """
    Set up the argument parser for the different commands.

    Return:
    Configured ArgumentParser object.
    """
    argument_parser = argparse.ArgumentParser(
        description='Build source code libraries from modules.')
    argument_parser.add_argument('-r', '--repository',
        metavar="REPO",
        dest='repositories',
        action='append',
        default=[],
        help="Repository file(s) which should be available for the current library. "
             "The loading of repository files from a VCS is only supported through "
             "the library configuration file.")
    argument_parser.add_argument('-c', '--config',
        dest='config',
        default='project.xml',
        help="Project configuration file. "
             "Specifies the required repositories, modules and options "
             "(default: '%(default)s').")
    argument_parser.add_argument('-p', '--path',
        dest='path',
        default='.',
        help="Path in which the library will be generated (default: '%(default)s').")
    argument_parser.add_argument('-D', '--option',
        metavar='OPTION',
        dest='options',
        action='append',
        type=str,
        default=[],
        help="Additional options. Options given here will be merged with options "
             "from the configuration file and will overwrite the configuration "
             "file definitions.")
    argument_parser.add_argument('-v', '--verbose',
        action='count',
        default=0,
        dest='verbose')
    argument_parser.add_argument("--plain",
        dest="plain",
        action="store_true",
        default=False,
        help="Disable styled output, only output plain ASCII.")
    argument_parser.add_argument('--version',
        action='version',
        version='%(prog)s {}'.format(__version__),
        help="Print the lbuild version number and exit.")

    subparsers = argument_parser.add_subparsers(title="Actions",
        dest="action")

    actions = [
        DiscoverAction(),
        DiscoverOptionsAction(),
        ValidateAction(),
        BuildAction(),
        CleanAction(),

        InitAction(),
        UpdateAction(),
        DependenciesAction(),
    ]
    for action in actions:
        action.register(subparsers)

    return argument_parser


def run(args, system_config=None):
    lbuild.logger.configure_logger(args.verbose)
    lbuild.format.plain = args.plain

    try:
        command = args.execute_action
    except AttributeError:
        raise lbuild.exception.LbuildArgumentException("No command specified")

    fail_silent = not command.__self__.config_required
    config = lbuild.config.ConfigNode.from_file(args.config, fail_silent=fail_silent)
    config.extend_last(system_config)
    return command(args, config)


def main():
    """
    Main entry point of lbuild.
    """
    try:
        argument_parser = prepare_argument_parser()

        commandline_arguments = sys.argv[1:]
        args = argument_parser.parse_args(commandline_arguments)
        lbuild.logger.configure_logger(args.verbose)

        output = run(args, lbuild.config.ConfigNode.from_filesystem())
        print(output)
    except lbuild.exception.LbuildAggregateException as aggregate:
        for error in aggregate.exceptions:
            sys.stderr.write('\nERROR: %s\n' % error)
        sys.exit(2)
    except lbuild.exception.LbuildForwardException as error:
        sys.stderr.write("\nERROR in '{}'\n".format(error.module))
        traceback.print_exception(type(error.exception),
                                  error.exception,
                                  error.exception.__traceback__,
                                  limit=-1)
        sys.exit(4)
    except lbuild.exception.LbuildArgumentException as error:
        argument_parser.print_help()
        print(error)
        sys.exit(2)
    except lbuild.exception.LbuildTemplateException as error:
        sys.stderr.write('\nERROR: %s\n' % error)
        traceback.print_exc()
        sys.exit(3)
    except lbuild.exception.LbuildException as error:
        sys.stderr.write('\nERROR: %s\n' % error)
        if args.verbose >= 2:
            traceback.print_exc()
        sys.exit(1)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2018, Niklas Hauser
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

import os
import re
import shutil
import colorful

import lbuild.node
import lbuild.filter

plain = False
WIDTH = shutil.get_terminal_size((200, 0)).columns

color_scheme = {
    "parser": None,
    "repository": None, #colorful.deepSkyBlue,
    "option": None, #colorful.lightSteelBlue,
    "module": None,
    "description": None,
    "short_description": None,
    "error": colorful.red,
}

def ansi_escape(obj=None):
    if plain:
        return ""
    name = obj
    if isinstance(obj, lbuild.node.BaseNode):
        name = obj.type.name.lower()
    col = color_scheme.get(name, "nope")
    if col == "nope":
        try:
            col = getattr(colorful, name)
        except:
            col = None
    return str(col) if col is not None else ""


class color_str:
    def __init__(self, string=None):
        self._string = ""
        self._content = []
        if isinstance(string, color_str):
            self._string = string._string
            self._content = string._content.copy()
        elif string is not None:
            self._string = string
            self._content = [("", self._string)]

    def wrap(self, name):
        style = ansi_escape(name)
        if style is not None:
            if name in ["underlined", "bold"]:
                if name == "underlined": name = "underline";
                close = ansi_escape("no_" + name)
            else:
                close = ansi_escape("close_fg_color")
            self._content = \
                [(style, "")] + \
                self._content + \
                [(close, "")]
        return self

    def join(self, strings):
        result = color_str()
        for ii, v in enumerate(strings):
            result += color_str(v)
            if ii < len(strings) - 1:
                result += self
        return result

    def limit(self, offset=0):
        if len(self) <= (WIDTH - offset):
            return str(self)

        mark = " ..."
        width = WIDTH - offset - len(mark)
        if width < 0: width = 0
        style_str = ""
        raw_string = ""
        for style, content in self._content:
            style_str += style;
            if len(raw_string + content) >= width:
                style_str += content[:width - len(raw_string)]
                return style_str + ansi_escape("reset") + mark
            raw_string += content
            style_str += content
        return style_str

    def __add__(self, other):
        color = color_str()
        if isinstance(other, color_str):
            color._content = self._content + other._content
            color._string = self._string + other._string
        else:
            color._string = other
            color._content = self._content
        return color

    def __len__(self):
        return len(self._string)

    def __iter__(self):
        return iter(self._string)

    def __contains__(self, string):
        return (string in self._string)

    def __getattr__(self, name):
        color = ansi_escape(name)
        if color is not None:
            self._content += [(color, "")]
        return self

    def __eq__(self, other):
        return self._string == other._string

    def __ne__(self, other):
        return self._string != other._string

    def __str__(self):
        string = ""
        for style, content in self._content:
            string += style + content
        return string + ansi_escape("reset")


c = color_str


def format_option_name(node, fullname=True):
    line = c(node.fullname if fullname else node.name).wrap(node).wrap("bold")
    if not node.is_default():
        line.wrap("underlined")
    return line


def format_option_value(node, single_line=True):
    offset = -1
    if node.value is None:
        return (c("REQUIRED").wrap("error").wrap("bold"), offset, False)
    value = node.format_value()
    if single_line:
        value = value.replace("\n", " ")
    return (c(value).wrap("bold"), offset, True)


def format_option_values(node, offset=0, single_line=True):
    values = c(node.format_values())
    if not single_line:
        if WIDTH - offset > 10:
            values = lbuild.filter.indent(lbuild.filter.wordwrap(values._string, WIDTH - offset), offset)
    return c(values).wrap("bold")


def format_option_value_description(node, offset=0, single_line=None):
    value = format_option_value(node, bool(single_line))
    offset = offset + len(value[0]) if value[1] < 0 else value[1]
    single_line = value[2] if single_line is None else single_line
    values = format_option_values(node, offset + 5, single_line)
    return value[0] + c(" in [") + values +  c("]")


def format_option_short_description(node):
    line = format_option_name(node) + c(" = ")
    line += format_option_value_description(node, offset=len(line))
    if "REQUIRED" not in line:
        line = line.limit()
    return str(line)


def format_description(node, description):
    output = [c(">> ") + c(node.fullname).wrap(node).wrap("bold"), c("")]
    if len(description):
        output += [c(description.strip(" \n"))]

    if node.type == node.Type.OPTION:
        value = format_option_value(node, single_line=False)[0]
        values = format_option_values(node, offset=9, single_line=False)
        output += [c(""), c("Value: ") + value, c("Inputs: [") + values + c("]")]

    for option in node.options:
        output.extend([c("\n"), c(">>") + c(option.description)])
    return "\n".join(map(str, output))


def format_short_description(node, description):
    lines = description.strip(" \n").splitlines() + [""]
    return lines[0].strip()


def format_node(node, prefix):
    context = c(node.fullname)
    if node._type == node.Type.REPOSITORY:
        context = c(node.name + " @ " + os.path.relpath(node._filepath, os.getcwd()))
    elif node._type == node.Type.OPTION:
        context = format_option_name(node, fullname=False)

    descr = (c(node.__class__.__name__ + "(") + context + c(")")).wrap(node)

    offset = node.depth*4
    if node._type == node.Type.OPTION:
        descr += c(" = ")
        descr += format_option_value_description(node, offset=offset+len(descr), single_line=True)
    else:
        descr += c("   " + node.short_description)

    return descr.limit(offset)
